@@ -1,10 +1,10 @@
 /*  GNU SED, a batch stream editor.
-    Copyright (C) 1989,90,91,92,93,94,95,98,99,2002,2003
+    Copyright (C) 1989,90,91,92,93,94,95,98,99,2002,2003,2004,2005,2006
     Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2, or (at your option)
+    the Free Software Foundation; either version 3, or (at your option)
     any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -14,7 +14,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. */
 
 /* compile.c: translate sed source into internal form */
 
@@ -134,7 +134,7 @@ static struct obstack obs;
 static const char errors[] =
   "multiple `!'s\0"
   "unexpected `,'\0"
-  "cannot use +N or ~N as first address\0"
+  "invalid usage of +N or ~N as first address\0"
   "unmatched `{'\0"
   "unexpected `}'\0"
   "extra characters after command\0"
@@ -159,22 +159,22 @@ static const char errors[] =
   "unknown command: `%c'";
 
 #define BAD_BANG (errors)
-#define BAD_COMMA (BAD_BANG + sizeof(N_("Multiple `!'s")))
-#define BAD_PLUS (BAD_COMMA + sizeof(N_("Unexpected `,'")))
-#define EXCESS_OPEN_BRACE (BAD_PLUS + sizeof(N_("Cannot use +N or ~N as first address")))
-#define EXCESS_CLOSE_BRACE (EXCESS_OPEN_BRACE + sizeof(N_("Unmatched `{'")))
-#define EXCESS_JUNK (EXCESS_CLOSE_BRACE + sizeof(N_("Unexpected `}'")))
-#define EXPECTED_SLASH (EXCESS_JUNK + sizeof(N_("Extra characters after command")))
-#define NO_CLOSE_BRACE_ADDR (EXPECTED_SLASH + sizeof(N_("Expected \\ after `a', `c' or `i'")))
+#define BAD_COMMA (BAD_BANG + sizeof(N_("multiple `!'s")))
+#define BAD_STEP (BAD_COMMA + sizeof(N_("unexpected `,'")))
+#define EXCESS_OPEN_BRACE (BAD_STEP + sizeof(N_("invalid usage of +N or ~N as first address")))
+#define EXCESS_CLOSE_BRACE (EXCESS_OPEN_BRACE + sizeof(N_("unmatched `{'")))
+#define EXCESS_JUNK (EXCESS_CLOSE_BRACE + sizeof(N_("unexpected `}'")))
+#define EXPECTED_SLASH (EXCESS_JUNK + sizeof(N_("extra characters after command")))
+#define NO_CLOSE_BRACE_ADDR (EXPECTED_SLASH + sizeof(N_("expected \\ after `a', `c' or `i'")))
 #define NO_COLON_ADDR (NO_CLOSE_BRACE_ADDR + sizeof(N_("`}' doesn't want any addresses")))
 #define NO_SHARP_ADDR (NO_COLON_ADDR + sizeof(N_(": doesn't want any addresses")))
-#define NO_COMMAND (NO_SHARP_ADDR + sizeof(N_("Comments don't accept any addresses")))
-#define ONE_ADDR (NO_COMMAND + sizeof(N_("Missing command")))
-#define UNTERM_ADDR_RE (ONE_ADDR + sizeof(N_("Command only uses one address")))
-#define UNTERM_S_CMD (UNTERM_ADDR_RE + sizeof(N_("Unterminated address regex")))
-#define UNTERM_Y_CMD (UNTERM_S_CMD + sizeof(N_("Unterminated `s' command")))
-#define UNKNOWN_S_OPT (UNTERM_Y_CMD + sizeof(N_("Unterminated `y' command")))
-#define EXCESS_P_OPT (UNKNOWN_S_OPT + sizeof(N_("Unknown option to `s'")))
+#define NO_COMMAND (NO_SHARP_ADDR + sizeof(N_("comments don't accept any addresses")))
+#define ONE_ADDR (NO_COMMAND + sizeof(N_("missing command")))
+#define UNTERM_ADDR_RE (ONE_ADDR + sizeof(N_("command only uses one address")))
+#define UNTERM_S_CMD (UNTERM_ADDR_RE + sizeof(N_("unterminated address regex")))
+#define UNTERM_Y_CMD (UNTERM_S_CMD + sizeof(N_("unterminated `s' command")))
+#define UNKNOWN_S_OPT (UNTERM_Y_CMD + sizeof(N_("unterminated `y' command")))
+#define EXCESS_P_OPT (UNKNOWN_S_OPT + sizeof(N_("unknown option to `s'")))
 #define EXCESS_G_OPT (EXCESS_P_OPT + sizeof(N_("multiple `p' options to `s' command")))
 #define EXCESS_N_OPT (EXCESS_G_OPT + sizeof(N_("multiple `g' options to `s' command")))
 #define ZERO_N_OPT (EXCESS_N_OPT + sizeof(N_("multiple number options to `s' command")))
@@ -195,7 +195,7 @@ bad_command(ch)
   char ch;
 {
   const char *msg = _(UNKNOWN_CMD);
-  char *unknown_cmd = xmalloc(strlen(msg) - 1);
+  char *unknown_cmd = xmalloc(strlen(msg));
   sprintf(unknown_cmd, msg, ch);
   bad_prog(unknown_cmd);
 }
@@ -255,7 +255,7 @@ savchar(ch)
   if (prog.cur)
     {
       if (prog.cur <= prog.base || *--prog.cur != ch)
-	panic(_("Called savchar() with unexpected pushback (%x)"),
+	panic("Called savchar() with unexpected pushback (%x)",
 	      CAST(unsigned char)ch);
     }
   else
@@ -374,17 +374,16 @@ read_filename()
   return b;
 }
 
-static struct output *get_openfile P_((struct output **file_ptrs, char *mode, bool fail));
+static struct output *get_openfile P_((struct output **file_ptrs, const char *mode, bool fail));
 static struct output *
 get_openfile(file_ptrs, mode, fail)
      struct output **file_ptrs;
-     char *mode;
+     const char *mode;
      bool fail;
 {
   struct buffer *b;
   char *file_name;
   struct output *p;
-  int is_stderr;
 
   b = read_filename();
   file_name = get_buffer(b);
@@ -441,7 +440,7 @@ next_cmd_entry(vectorp)
   cmd = v->v + v->v_length;
   cmd->a1 = NULL;
   cmd->a2 = NULL;
-  cmd->a1_matched = false;
+  cmd->range_state = RANGE_INACTIVE;
   cmd->addr_bang = false;
   cmd->cmd = '\0';	/* something invalid, to catch bugs early */
 
@@ -463,7 +462,7 @@ snarf_char_class(b, cur_stat)
   ch = inchar();
   if (ch == '^')
     ch = add_then_next(b, ch);
-  else if (ch == CLOSE_BRACKET)
+  if (ch == CLOSE_BRACKET)
     ch = add_then_next(b, ch);
   
   /* States are:
@@ -511,7 +510,7 @@ snarf_char_class(b, cur_stat)
 	  if (pending_mb)
 	    continue;
 
-	  if (state == 0)
+	  if (state == 0 || state == 1)
 	    return ch;
 	  else if (state == 3)
 	    state = 0;
@@ -588,8 +587,8 @@ match_slash(slash, regex)
   return NULL;
 }
 
-static bool mark_subst_opts P_((struct subst *cmd));
-static bool
+static int mark_subst_opts P_((struct subst *cmd));
+static int
 mark_subst_opts(cmd)
   struct subst *cmd;
 {
@@ -1049,7 +1048,7 @@ compile_program(vector)
 	{
 	  if (a.addr_type == ADDR_IS_STEP
 	      || a.addr_type == ADDR_IS_STEP_MOD)
-	    bad_prog(_(BAD_PLUS));
+	    bad_prog(_(BAD_STEP));
 
 	  cur_cmd->a1 = MEMDUP(&a, 1, struct addr);
 	  ch = in_nonblank();
@@ -1057,20 +1056,15 @@ compile_program(vector)
 	    {
 	      if (!compile_address(&a, in_nonblank()))
 		bad_prog(_(BAD_COMMA));
-	      if (a.addr_type == ADDR_IS_NUM)
-		a.addr_type = ADDR_IS_NUM2;
+
 	      cur_cmd->a2 = MEMDUP(&a, 1, struct addr);
 	      ch = in_nonblank();
-
-	      if ((cur_cmd->a1->addr_type == ADDR_IS_NUM
-	           || cur_cmd->a1->addr_type == ADDR_IS_NUM_MOD)
-	          && cur_cmd->a1->addr_number == 0)
-	        cur_cmd->a1_matched = true;
 	    }
 
-	  if (cur_cmd->a1->addr_type == ADDR_IS_NUM
-	      && cur_cmd->a1->addr_number == 0
-	      && (!cur_cmd->a2 || cur_cmd->a2->addr_type != ADDR_IS_REGEX))
+	  if ((cur_cmd->a1->addr_type == ADDR_IS_NUM
+	       && cur_cmd->a1->addr_number == 0)
+	      && ((!cur_cmd->a2 || cur_cmd->a2->addr_type != ADDR_IS_REGEX)
+		  || posixicity == POSIXLY_BASIC))
 	    bad_prog(_(INVALID_LINE_0));
 	}
       if (ch == '!')
@@ -1240,7 +1234,7 @@ compile_program(vector)
 	  break;
 
         case 'R':
-	  cur_cmd->x.fp = get_openfile(&file_read, "r", false)->fp;
+	  cur_cmd->x.fp = get_openfile(&file_read, read_mode, false)->fp;
 	  break;
 
         case 'W':
@@ -1267,7 +1261,7 @@ compile_program(vector)
 
 	    flags = mark_subst_opts(cur_cmd->x.cmd_subst);
 	    cur_cmd->x.cmd_subst->regx =
-	      compile_regex(b, flags, cur_cmd->x.cmd_subst->max_id);
+	      compile_regex(b, flags, cur_cmd->x.cmd_subst->max_id + 1);
 	    free_buffer(b);
 	  }
 	  break;
@@ -1280,12 +1274,12 @@ compile_program(vector)
             char *src_buf, *dest_buf;
   
 	    slash = inchar();
-	    if ( !(b = match_slash(slash, true)) )
+	    if ( !(b = match_slash(slash, false)) )
 	      bad_prog(_(UNTERM_Y_CMD));
             src_buf = get_buffer(b);
 	    len = normalize_text(src_buf, size_buffer (b), TEXT_BUFFER);
 
-            if ( !(b2 = match_slash(slash, true)) )
+            if ( !(b2 = match_slash(slash, false)) )
  	      bad_prog(_(UNTERM_Y_CMD));
             dest_buf = get_buffer(b2);
 	    dest_len = normalize_text(dest_buf, size_buffer (b2), TEXT_BUFFER);
@@ -1324,7 +1318,7 @@ compile_program(vector)
                 cur_cmd->x.translatemb = trans_pairs;
                 for (i = 0; i < src_char_num; i++)
                   {
-                    if (idx >= len)
+                    if (idx >= dest_len)
                       bad_prog(_(Y_CMD_LEN));
 
                     /* Set the i-th source character.  */
@@ -1334,7 +1328,7 @@ compile_program(vector)
                     src_buf += src_lens[i]; /* Forward to next character.  */
 
                     /* Fetch the i-th destination character.  */
-                    mbclen = MBRLEN (dest_buf + idx, len - idx, &cur_stat);
+                    mbclen = MBRLEN (dest_buf + idx, dest_len - idx, &cur_stat);
                     /* An invalid sequence, or a truncated multibyte character.
                        We treat it as a singlebyte character.  */
                     if (mbclen == (size_t) -1 || mbclen == (size_t) -2
@@ -1348,12 +1342,13 @@ compile_program(vector)
                     idx += mbclen; /* Forward to next character.  */
                   }
                 trans_pairs[2 * i] = NULL;
-                if (idx != len)
+                if (idx != dest_len)
                   bad_prog(_(Y_CMD_LEN));
               }
             else
               {
-	        char *translate = OB_MALLOC(&obs, YMAP_LENGTH, char);
+	        unsigned char *translate =
+		  OB_MALLOC(&obs, YMAP_LENGTH, unsigned char);
                 unsigned char *ustring = CAST(unsigned char *)src_buf;
 
 		if (len != dest_len)
@@ -1363,7 +1358,7 @@ compile_program(vector)
 	          translate[len] = len;
 
                 while (dest_len--)
-                  translate[(unsigned char)*ustring++] = *dest_buf++;
+                  translate[*ustring++] = (unsigned char)*dest_buf++;
 
 	        cur_cmd->x.translate = translate;
 	      }
@@ -1587,7 +1582,6 @@ compile_file(cur_program, cmdfile)
   struct vector *cur_program;
   const char *cmdfile;
 {
-  size_t len;
   struct vector *ret;
 
   prog.file = stdin;
@@ -1629,8 +1623,9 @@ check_final_program(program)
   if (pending_text)
     {
       old_text_buf->text_length = size_buffer(pending_text);
-      old_text_buf->text = MEMDUP(get_buffer(pending_text),
-				  old_text_buf->text_length, char);
+      if (old_text_buf->text_length)
+        old_text_buf->text = MEMDUP(get_buffer(pending_text),
+				    old_text_buf->text_length, char);
       free_buffer(pending_text);
       pending_text = NULL;
     }
@@ -1647,7 +1642,7 @@ check_final_program(program)
       else
 	{
 	  if (*go->name)
-	    panic(_("Can't find label for jump to `%s'"), go->name);
+	    panic(_("can't find label for jump to `%s'"), go->name);
 	  program->v[go->v_index].x.jump_index = program->v_length;
 	}
     }
