@@ -1,11 +1,10 @@
-
 /*  GNU SED, a batch stream editor.
-    Copyright (C) 1989,90,91,92,93,94,95,98,99,2002,2003
+    Copyright (C) 1989,90,91,92,93,94,95,98,99,2002,2003,2004,2005,2006
     Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2, or (at your option)
+    the Free Software Foundation; either version 3, or (at your option)
     any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -15,7 +14,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. */
 
 #undef EXPERIMENTAL_DASH_N_OPTIMIZATION	/*don't use -- is very buggy*/
 #define INITIAL_BUFFER_SIZE	50
@@ -23,6 +22,7 @@
 
 #include "sed.h"
 
+#include <stddef.h>
 #include <stdio.h>
 #include <ctype.h>
 
@@ -35,6 +35,8 @@ extern int errno;
 # include <unistd.h>
 #endif
 
+#include "acl.h"
+
 #ifdef __GNUC__
 # if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__-0 >= 7)
    /* silence warning about unused parameter even for "gcc -W -Wunused" */
@@ -71,6 +73,7 @@ extern int errno;
 #endif
 
 #include <sys/stat.h>
+#include "stat-macros.h"
 
 
 /* Sed operates a line at a time. */
@@ -85,6 +88,12 @@ struct line {
 #endif
 };
 
+#ifdef HAVE_MBRTOWC
+#define SIZEOF_LINE	offsetof (struct line, mbstate)
+#else
+#define SIZEOF_LINE	(sizeof (struct line))
+#endif
+
 /* A queue of text to write out at the end of a cycle
    (filled by the "a", "r" and "R" commands.) */
 struct append_queue {
@@ -97,21 +106,34 @@ struct append_queue {
 
 /* State information for the input stream. */
 struct input {
-  char **file_list;	/* The list of yet-to-be-opened files.
-			   It is invalid for file_list to be NULL.
-			   When *file_list is NULL we are
-			   currently processing the last file. */
-  countT bad_count;	/* count of files we failed to open */
-  countT line_number;	/* current input line number (over all files) */
+  /* The list of yet-to-be-opened files.  It is invalid for file_list
+     to be NULL.  When *file_list is NULL we are currently processing
+     the last file.  */
 
+  char **file_list;
+
+  /* Count of files we failed to open. */
+  countT bad_count;	    
+
+  /* Current input line number (over all files).  */
+  countT line_number;	    
+
+  /* True if we'll reset line numbers and addresses before
+     starting to process the next (possibly the first) file.  */
+  bool reset_at_next_file;
+
+  /* Function to read one line.  If FP is NULL, read_fn better not
+     be one which uses fp; in particular, read_always_fail() is
+     recommended. */
   bool (*read_fn) P_((struct input *));	/* read one line */
-  /* If fp is NULL, read_fn better not be one which uses fp;
-     in particular, read_always_fail() is recommended. */
 
   char *out_file_name;
+
   const char *in_file_name;
 
-  FILE *fp;		/* if NULL, none of the following are valid */
+  /* if NULL, none of the following are valid */
+  FILE *fp;
+
   bool no_buffering;
 };
 
@@ -217,25 +239,26 @@ str_append(to, string, length)
   to->length = new_length;
 
 #ifdef HAVE_MBRTOWC
-  if (mb_cur_max == 1)
-    return;
+  if (mb_cur_max > 1 && !is_utf8)
+    while (length)
+      {
+        size_t n = MBRLEN (string, length, &to->mbstate);
 
-  while (length)
-    {
-      int n = MBRLEN (string, length, &to->mbstate);
+        /* An invalid sequence is treated like a singlebyte character. */
+        if (n == (size_t) -1)
+	  {
+	    memset (&to->mbstate, 0, sizeof (to->mbstate));
+	    n = 1;
+	  }
 
-      /* An invalid sequence is treated like a singlebyte character. */
-      if (n == -1)
-	{
-	  memset (&to->mbstate, 0, sizeof (to->mbstate));
-	  n = 1;
-	}
-
-      if (n > 0)
-	length -= n;
-      else
-	break;
-    }
+        if (n > 0)
+	  {
+	    string += n;
+	    length -= n;
+	  }
+        else
+	  break;
+      }
 #endif
 }
 
@@ -248,79 +271,76 @@ str_append_modified(to, string, length, type)
   size_t length;
   enum replacement_types type;
 {
+#ifdef HAVE_MBRTOWC
+  mbstate_t from_stat;
+
+  if (type == REPL_ASIS)
+    {
+      str_append(to, string, length);
+      return;
+    }
+
+  if (to->alloc - to->length < length * mb_cur_max)
+    resize_line(to, to->length + length * mb_cur_max);
+
+  MEMCPY (&from_stat, &to->mbstate, sizeof(mbstate_t));
+  while (length)
+    {
+      wchar_t wc;
+      int n = MBRTOWC (&wc, string, length, &from_stat);
+
+      /* An invalid sequence is treated like a singlebyte character. */
+      if (n == -1)
+        {
+          memset (&to->mbstate, 0, sizeof (from_stat));
+          n = 1;
+        }
+
+      if (n > 0)
+        string += n, length -= n;
+      else
+	{
+	  /* Incomplete sequence, copy it manually.  */
+	  str_append(to, string, length);
+	  return;
+	}
+
+      /* Convert the first character specially... */
+      if (type & (REPL_UPPERCASE_FIRST | REPL_LOWERCASE_FIRST))
+	{
+          if (type & REPL_UPPERCASE_FIRST)
+            wc = towupper(wc);
+          else
+            wc = towlower(wc);
+
+          type &= ~(REPL_LOWERCASE_FIRST | REPL_UPPERCASE_FIRST);
+	  if (type == REPL_ASIS)
+	    {
+	      n = WCRTOMB (to->active + to->length, wc, &to->mbstate);
+	      to->length += n;
+	      str_append(to, string, length);
+	      return;
+	    }
+        }
+
+      else if (type & REPL_UPPERCASE)
+        wc = towupper(wc);
+      else
+        wc = towlower(wc);
+
+      /* Copy the new wide character to the end of the string. */
+      n = WCRTOMB (to->active + to->length, wc, &to->mbstate);
+      to->length += n;
+      if (n == -1)
+	{
+	  fprintf (stderr, "Case conversion produced an invalid character!");
+	  abort ();
+	}
+    }
+#else
   size_t old_length = to->length;
   char *start, *end;
 
-  if (length == 0)
-    return;
-
-#ifdef HAVE_MBRTOWC
-  {
-    mbstate_t from_stat;
-
-    if (type == REPL_ASIS)
-      {
-	str_append(to, string, length);
-        return;
-      }
-
-    if (to->alloc - to->length < length * mb_cur_max)
-      resize_line(to, to->length + length * mb_cur_max);
-
-    MEMCPY (&from_stat, &to->mbstate, sizeof(mbstate_t));
-    while (length)
-      {
-	wchar_t wc;
-        int n = MBRTOWC (&wc, string, length, &from_stat);
-
-        /* An invalid sequence is treated like a singlebyte character. */
-        if (n == -1)
-          {
-            memset (&to->mbstate, 0, sizeof (from_stat));
-            n = 1;
-          }
-
-        if (n > 0)
-          string += n, length -= n;
-        else
-	  {
-	    /* Incomplete sequence, copy it manually.  */
-	    str_append(to, string, length);
-	    return;
-	  }
-
-	/* Convert the first character specially... */
-        if (type & (REPL_UPPERCASE_FIRST | REPL_LOWERCASE_FIRST))
-	  {
-            if (type & REPL_UPPERCASE_FIRST)
-              wc = towupper(wc);
-            else
-              wc = towlower(wc);
-
-            type &= ~(REPL_LOWERCASE_FIRST | REPL_UPPERCASE_FIRST);
-	    if (type == REPL_ASIS)
-	      {
-		str_append(to, string, length);
-	        return;
-	      }
-          }
-
-        else if (type & REPL_UPPERCASE)
-          wc = towupper(wc);
-        else
-          wc = towlower(wc);
-
-	/* Copy the new wide character to the end of the string. */
-	n = wcrtomb (to->active + to->length, wc, &to->mbstate);
-        to->length += n;
-	if (n == -1)
-	  {
-	    fprintf (stderr, "Case conversion produced an invalid character!");
-	    abort ();
-	  }
-      }
-  }
-#else
   str_append(to, string, length);
   start = to->active + old_length;
   end = start + length;
@@ -352,11 +372,13 @@ str_append_modified(to, string, length, type)
 #endif
 }
 
-/* initialize a "struct line" buffer */
-static void line_init P_((struct line *, size_t initial_size));
+/* Initialize a "struct line" buffer.  Copy multibyte state from `state'
+   if not null.  */
+static void line_init P_((struct line *, struct line *, size_t initial_size));
 static void
-line_init(buf, initial_size)
+line_init(buf, state, initial_size)
   struct line *buf;
+  struct line *state;
   size_t initial_size;
 {
   buf->text = MALLOC(initial_size, char);
@@ -366,18 +388,43 @@ line_init(buf, initial_size)
   buf->chomped = true;
 
 #ifdef HAVE_MBRTOWC
-  memset (&buf->mbstate, 0, sizeof (buf->mbstate));
+  if (state)
+    memcpy (&buf->mbstate, &state->mbstate, sizeof (buf->mbstate));
+  else
+    memset (&buf->mbstate, 0, sizeof (buf->mbstate));
 #endif
+}
 
+/* Reset a "struct line" buffer to length zero.  Copy multibyte state from
+   `state' if not null.  */
+static void line_reset P_((struct line *, struct line *));
+static void
+line_reset(buf, state)
+  struct line *buf, *state;
+{
+  if (buf->alloc == 0)
+    line_init(buf, state, INITIAL_BUFFER_SIZE);
+  else
+    {
+      buf->length = 0;
+#ifdef HAVE_MBRTOWC
+      if (state)
+        memcpy (&buf->mbstate, &state->mbstate, sizeof (buf->mbstate));
+      else
+        memset (&buf->mbstate, 0, sizeof (buf->mbstate));
+#endif
+    }
 }
 
 /* Copy the contents of the line `from' into the line `to'.
-   This destroys the old contents of `to'. */
-static void line_copy P_((struct line *from, struct line *to));
+   This destroys the old contents of `to'.
+   Copy the multibyte state if `state' is true. */
+static void line_copy P_((struct line *from, struct line *to, bool state));
 static void
-line_copy(from, to)
+line_copy(from, to, state)
   struct line *from;
   struct line *to;
+  bool state;
 {
   /* Remove the inactive portion in the destination buffer. */
   to->alloc += to->active - to->text;
@@ -401,38 +448,53 @@ line_copy(from, to)
   MEMCPY(to->active, from->active, from->length);
 
 #ifdef HAVE_MBRTOWC
-  MEMCPY(&to->mbstate, &from->mbstate, sizeof (from->mbstate));
+  if (state)
+    MEMCPY(&to->mbstate, &from->mbstate, sizeof (from->mbstate));
 #endif
 }
 
-/* Append the contents of the line `from' to the line `to'. */
-static void line_append P_((struct line *from, struct line *to));
+/* Append the contents of the line `from' to the line `to'.
+   Copy the multibyte state if `state' is true. */
+static void line_append P_((struct line *from, struct line *to, bool state));
 static void
-line_append(from, to)
+line_append(from, to, state)
   struct line *from;
   struct line *to;
+  bool state;
 {
   str_append(to, "\n", 1);
   str_append(to, from->active, from->length);
   to->chomped = from->chomped;
 
 #ifdef HAVE_MBRTOWC
-  memcpy(&to->mbstate, &from->mbstate, sizeof (from->mbstate));
+  if (state)
+    MEMCPY (&to->mbstate, &from->mbstate, sizeof (from->mbstate));
 #endif
 }
 
-/* Exchange the contents of two "struct line" buffers. */
-static void line_exchange P_((struct line *, struct line *));
+/* Exchange two "struct line" buffers.
+   Copy the multibyte state if `state' is true. */
+static void line_exchange P_((struct line *a, struct line *b, bool state));
 static void
-line_exchange(a, b)
+line_exchange(a, b, state)
   struct line *a;
   struct line *b;
+  bool state;
 {
   struct line t;
 
-  MEMCPY(&t,  a, sizeof(struct line));
-  MEMCPY( a,  b, sizeof(struct line));
-  MEMCPY( b, &t, sizeof(struct line));
+  if (state)
+    {
+      MEMCPY(&t,  a, sizeof (struct line));
+      MEMCPY( a,  b, sizeof (struct line));
+      MEMCPY( b, &t, sizeof (struct line));
+    }
+  else
+    {
+      MEMCPY(&t,  a, SIZEOF_LINE);
+      MEMCPY( a,  b, SIZEOF_LINE);
+      MEMCPY( b, &t, SIZEOF_LINE);
+    }
 }
 
 
@@ -453,15 +515,13 @@ read_file_line(input)
   static char *b;
   static size_t blen;
 
-  long result = getline (&b, &blen, input->fp);
-
+  long result = ck_getline (&b, &blen, input->fp);
   if (result <= 0)
     return false;
 
   /* Remove the trailing new-line that is left by getline. */
-  else if (b[result - 1] == '\n')
+  if (b[result - 1] == '\n')
     --result;
-
   else
     line.chomped = false;
 
@@ -499,11 +559,12 @@ output_line(text, length, nl, outf)
   bool nl;
   struct output *outf;
 {
-  output_missing_newline(outf);
+  if (!text)
+    return;
 
+  output_missing_newline(outf);
   if (length)
     ck_fwrite(text, 1, length, outf->fp);
-
   if (nl)
     ck_fwrite("\n", 1, 1, outf->fp);
   else
@@ -570,7 +631,7 @@ dump_append_queue()
 	     be treated as if it were an empty file, causing no error
 	     condition."  IEEE Std 1003.2-1992
 	     So, don't fail. */
-	  fp = ck_fopen(p->fname, "r", false);
+	  fp = ck_fopen(p->fname, read_mode, false);
 	  if (fp)
 	    {
 	      while ((cnt = ck_fread(buf, 1, sizeof buf, fp)) > 0)
@@ -607,7 +668,7 @@ get_backup_file_name(name)
        asterisk = strchr(old_asterisk, '*');
        old_asterisk = asterisk + 1)
     {
-      memcpy (p, old_asterisk, asterisk - old_asterisk);
+      MEMCPY (p, old_asterisk, asterisk - old_asterisk);
       p += asterisk - old_asterisk;
       strcpy (p, name);
       p += name_length;
@@ -627,12 +688,12 @@ open_next_file(name, input)
 {
   buffer.length = 0;
 
-  if (name[0] == '-' && name[1] == '\0')
+  if (name[0] == '-' && name[1] == '\0' && !in_place_extension)
     {
       clearerr(stdin);	/* clear any stale EOF indication */
-      input->fp = stdin;
+      input->fp = ck_fdopen (fileno (stdin), "stdin", read_mode, false);
     }
-  else if ( ! (input->fp = ck_fopen(name, "r", false)) )
+  else if ( ! (input->fp = ck_fopen(name, read_mode, false)) )
     {
       const char *ptr = strerror(errno);
       fprintf(stderr, _("%s: can't read %s: %s\n"), myname, name, ptr);
@@ -645,7 +706,7 @@ open_next_file(name, input)
 
   if (in_place_extension)
     {
-      int output_fd;
+      int input_fd, output_fd;
       char *tmpdir = ck_strdup(name), *p;
       struct stat st;
 
@@ -656,27 +717,30 @@ open_next_file(name, input)
 	strcpy(tmpdir, ".");
       
       input->in_file_name = name;
-      input->out_file_name = temp_file_template (tmpdir, "sed");
-      output_fd = mkstemp (input->out_file_name);
+
+      if (isatty (fileno (input->fp)))
+        panic(_("couldn't edit %s: is a terminal"), input->in_file_name);
+
+      input_fd = fileno (input->fp);
+      fstat (input_fd, &st);
+      if (!S_ISREG (st.st_mode))
+        panic(_("couldn't edit %s: not a regular file"), input->in_file_name);
+
+      output_file.fp = ck_mkstemp (&input->out_file_name, tmpdir, "sed");
+      output_file.missing_newline = false;
       free (tmpdir);
 
-      if (output_fd == -1)
-        panic(_("Couldn't open temporary file %s: %s"), input->out_file_name, strerror(errno));
+      if (!output_file.fp)
+        panic(_("couldn't open temporary file %s: %s"), input->out_file_name, strerror(errno));
 
-      fstat (fileno (input->fp), &st);
-#ifdef HAVE_FCHMOD
-      fchmod (output_fd, st.st_mode);
-#endif
+      output_fd = fileno (output_file.fp);
 #ifdef HAVE_FCHOWN
       if (fchown (output_fd, st.st_uid, st.st_gid) == -1)
         fchown (output_fd, -1, st.st_gid);
 #endif
-
-      output_file.fp = fdopen (output_fd, "w");
-      output_file.missing_newline = false;
-
-      if (!output_file.fp)
-        panic(_("Couldn't open temporary file %s: %s"), input->out_file_name, strerror(errno));
+      copy_acl (input->in_file_name, input_fd,
+		input->out_file_name, output_fd,
+		st.st_mode);
     }
   else
     output_file.fp = stdout;
@@ -692,22 +756,25 @@ closedown(input)
   input->read_fn = read_always_fail;
   if (!input->fp)
     return;
-  if (input->fp != stdin) /* stdin can be reused on tty and tape devices */
-    ck_fclose(input->fp);
 
   if (in_place_extension && output_file.fp != NULL)
     {
-      int fd = fileno (output_file.fp);
+      const char *target_name;
+      ck_fclose (output_file.fp);
+
+      if (follow_symlinks)
+	target_name = follow_symlink (input->in_file_name);
+      else
+	target_name = input->in_file_name;
+
       if (strcmp(in_place_extension, "*") != 0)
         {
-          char *backup_file_name = get_backup_file_name(input->in_file_name);
-          rename (input->in_file_name, backup_file_name);
+          char *backup_file_name = get_backup_file_name(target_name);
+	  ck_rename (target_name, backup_file_name, input->out_file_name);
           free (backup_file_name);
-        }
+	}
 
-      ck_fclose (output_file.fp);
-      close (fd);
-      rename (input->out_file_name, input->in_file_name);
+      ck_rename (input->out_file_name, target_name, input->out_file_name);
       free (input->out_file_name);
     }
 
@@ -724,14 +791,12 @@ reset_addresses(vec)
   int n;
 
   for (cur_cmd = vec->v, n = vec->v_length; n--; cur_cmd++)
-    if (cur_cmd->a1)
-      {
-        enum addr_types type = cur_cmd->a1->addr_type;
-        cur_cmd->a1_matched = (type == ADDR_IS_NUM || type == ADDR_IS_NUM_MOD)
-			      && cur_cmd->a1->addr_number == 0;
-      }
+    if (cur_cmd->a1
+	&& cur_cmd->a1->addr_type == ADDR_IS_NUM
+	&& cur_cmd->a1->addr_number == 0)
+      cur_cmd->range_state = RANGE_ACTIVE;
     else
-      cur_cmd->a1_matched = false;
+      cur_cmd->range_state = RANGE_INACTIVE;
 }
 
 /* Read in the next line of input, and store it in the pattern space.
@@ -757,7 +822,7 @@ read_pattern_space(input, the_program, append)
       if (!*input->file_list)
 	return false;
 
-      if (separate_files)
+      if (input->reset_at_next_file)
 	{
 	  input->line_number = 0;
 	  reset_addresses (the_program);
@@ -768,6 +833,8 @@ read_pattern_space(input, the_program, append)
 	     we might still have to output the missing new-line.  */
 	  if (in_place_extension)
 	    output_file.missing_newline = false;
+
+	  input->reset_at_next_file = separate_files;
 	}
 
       open_next_file (*input->file_list++, input);
@@ -838,14 +905,10 @@ match_an_address_p(addr, input)
     case ADDR_IS_REGEX:
       return match_regex(addr->addr_regex, line.active, line.length, 0, NULL, 0);
 
-    case ADDR_IS_NUM:
-      return (addr->addr_number == input->line_number);
-
     case ADDR_IS_NUM_MOD:
       return (input->line_number >= addr->addr_number
 	      && ((input->line_number - addr->addr_number) % addr->addr_step) == 0);
 
-    case ADDR_IS_NUM2:
     case ADDR_IS_STEP:
     case ADDR_IS_STEP_MOD:
       /* reminder: these are only meaningful for a2 addresses */
@@ -856,8 +919,10 @@ match_an_address_p(addr, input)
     case ADDR_IS_LAST:
       return test_eof(input);
 
+      /* ADDR_IS_NUM is handled in match_address_p.  */
+    case ADDR_IS_NUM:
     default:
-      panic(_("INTERNAL ERROR: bad address type"));
+      panic("INTERNAL ERROR: bad address type");
     }
   /*NOTREACHED*/
   return false;
@@ -870,40 +935,76 @@ match_address_p(cmd, input)
   struct sed_cmd *cmd;
   struct input *input;
 {
-  bool addr_matched = cmd->a1_matched;
+  if (!cmd->a1)
+    return true;
 
-  if (addr_matched)
+  if (cmd->range_state != RANGE_ACTIVE)
     {
-      if (match_an_address_p(cmd->a2, input))
-	cmd->a1_matched = false;
-    }
-  else if (!cmd->a1 || match_an_address_p(cmd->a1, input))
-    {
-      addr_matched = true;
-      if (cmd->a2)
+      /* Find if we are going to activate a range.  Handle ADDR_IS_NUM
+	 specially: it represent an "absolute" state, it should not
+	 be computed like regexes.  */
+      if (cmd->a1->addr_type == ADDR_IS_NUM)
 	{
-	  cmd->a1_matched = true;
-	  switch (cmd->a2->addr_type)
-	    {
-	    case ADDR_IS_REGEX:
-	      break;
-	    case ADDR_IS_STEP:
-	      cmd->a2->addr_number = input->line_number + cmd->a2->addr_step;
-	      break;
-	    case ADDR_IS_STEP_MOD:
-	      cmd->a2->addr_number = input->line_number + cmd->a2->addr_step
-				     - (input->line_number%cmd->a2->addr_step);
-	      break;
-	    default:
-	      if (match_an_address_p(cmd->a2, input))
-		cmd->a1_matched = false;
-	      break;
-	    }
+	  if (!cmd->a2)
+	    return (input->line_number == cmd->a1->addr_number);
+
+	  if (cmd->range_state == RANGE_CLOSED
+	      || input->line_number < cmd->a1->addr_number)
+	    return false;
 	}
+      else
+	{
+          if (!cmd->a2)
+	    return match_an_address_p(cmd->a1, input);
+
+	  if (!match_an_address_p(cmd->a1, input))
+            return false;
+	}
+
+      /* Ok, start a new range.  */
+      cmd->range_state = RANGE_ACTIVE;
+      switch (cmd->a2->addr_type)
+	{
+	case ADDR_IS_REGEX:
+	  /* Always include at least two lines.  */
+	  return true;
+	case ADDR_IS_NUM:
+	  /* Same handling as below, but always include at least one line.  */
+          if (input->line_number >= cmd->a2->addr_number)
+	    cmd->range_state = RANGE_CLOSED;
+          return true;
+	case ADDR_IS_STEP:
+	  cmd->a2->addr_number = input->line_number + cmd->a2->addr_step;
+	  return true;
+	case ADDR_IS_STEP_MOD:
+	  cmd->a2->addr_number = input->line_number + cmd->a2->addr_step
+				 - (input->line_number%cmd->a2->addr_step);
+	  return true;
+	default:
+	  break;
+        }
     }
-  if (cmd->addr_bang)
-    return !addr_matched;
-  return addr_matched;
+
+  /* cmd->range_state == RANGE_ACTIVE.  Check if the range is
+     ending; also handle ADDR_IS_NUM specially in this case.  */
+
+  if (cmd->a2->addr_type == ADDR_IS_NUM)
+    {
+      /* If the second address is a line number, and if we got past
+         that line, fail to match (it can happen when you jump
+	 over such addresses with `b' and `t'.  Use RANGE_CLOSED
+         so that the range is not re-enabled anymore.  */
+      if (input->line_number >= cmd->a2->addr_number)
+	cmd->range_state = RANGE_CLOSED;
+
+      return (input->line_number <= cmd->a2->addr_number);
+   }
+
+  /* Other addresses are treated as usual.  */
+  if (match_an_address_p(cmd->a2, input))
+    cmd->range_state = RANGE_CLOSED;
+
+  return true;
 }
 
 
@@ -999,16 +1100,18 @@ append_replacement (buf, p, regs, repl_mod)
         }
 
       if (0 <= i)
-        if (regs->end[i] == regs->start[i] && p->repl_type & REPL_MODIFIERS)
-          /* Save this modifier, we shall apply it later.
-	     e.g. in s/()([a-z])/\u\1\2/
-	     the \u modifier is applied to \2, not \1 */
-	  repl_mod = curr_type & REPL_MODIFIERS;
+	{
+          if (regs->end[i] == regs->start[i] && p->repl_type & REPL_MODIFIERS)
+            /* Save this modifier, we shall apply it later.
+	       e.g. in s/()([a-z])/\u\1\2/
+	       the \u modifier is applied to \2, not \1 */
+	    repl_mod = curr_type & REPL_MODIFIERS;
 
-	else
-	  str_append_modified(buf, line.active + regs->start[i],
-			      CAST(size_t)(regs->end[i] - regs->start[i]),
-			      curr_type);
+	  else if (regs->end[i] != regs->start[i])
+	    str_append_modified(buf, line.active + regs->start[i],
+			        CAST(size_t)(regs->end[i] - regs->start[i]),
+			        curr_type);
+	}
     }
 
   return repl_mod;
@@ -1024,38 +1127,37 @@ do_subst(sub)
   countT count = 0;	/* number of matches found */
   bool again = true;
 
-#define MAX_BACKREFERENCES 10
   static struct re_registers regs;
 
-  if (s_accum.alloc == 0)
-    line_init(&s_accum, INITIAL_BUFFER_SIZE);
-  s_accum.length = 0;
+  line_reset(&s_accum, &line);
 
   /* The first part of the loop optimizes s/xxx// when xxx is at the
      start, and s/xxx$// */
   if (!match_regex(sub->regx, line.active, line.length, start,
-		   &regs, MAX_BACKREFERENCES))
+		   &regs, sub->max_id + 1))
     return;
   
   if (!sub->replacement && sub->numb <= 1)
-    if (regs.start[0] == 0 && !sub->global)
-      {
-	/* We found a match, set the `replaced' flag. */
-	replaced = true;
+    {
+      if (regs.start[0] == 0 && !sub->global)
+        {
+	  /* We found a match, set the `replaced' flag. */
+	  replaced = true;
 
-	line.active += regs.end[0];
-	line.length -= regs.end[0];
-	line.alloc -= regs.end[0];
-	goto post_subst;
-      }
-    else if (regs.end[0] == line.length)
-      {
-	/* We found a match, set the `replaced' flag. */
-	replaced = true;
+	  line.active += regs.end[0];
+	  line.length -= regs.end[0];
+	  line.alloc -= regs.end[0];
+	  goto post_subst;
+        }
+      else if (regs.end[0] == line.length)
+        {
+	  /* We found a match, set the `replaced' flag. */
+	  replaced = true;
 
-	line.length = regs.start[0];
-	goto post_subst;
-      }
+	  line.length = regs.start[0];
+	  goto post_subst;
+        }
+    }
 
   do
     {
@@ -1112,7 +1214,7 @@ do_subst(sub)
   while (again
 	 && start <= line.length
 	 && match_regex(sub->regx, line.active, line.length, start,
-			&regs, MAX_BACKREFERENCES));
+			&regs, sub->max_id + 1));
 
   /* Copy stuff to the right of the last match into the output string. */
   if (start < line.length)
@@ -1121,8 +1223,8 @@ do_subst(sub)
 
   /* Exchange line and s_accum.  This can be much cheaper
      than copying s_accum.active into line.text (for huge lines). */
-  line_exchange(&line, &s_accum);
-  
+  line_exchange(&line, &s_accum, false);
+
   /* Finish up. */
   if (count < sub->numb)
     return;
@@ -1135,7 +1237,7 @@ do_subst(sub)
     {
 #ifdef HAVE_POPEN
       FILE *pipe;
-      s_accum.length = 0;
+      line_reset(&s_accum, NULL);
       
       str_append (&line, "", 1);
       pipe = popen(line.active, "r");
@@ -1152,7 +1254,10 @@ do_subst(sub)
 	  
 	  pclose (pipe);
 
-	  line_exchange(&line, &s_accum);
+	  /* Exchange line and s_accum.  This can be much cheaper than copying
+	     s_accum.active into line.text (for huge lines).  See comment above
+	     for 'g' as to while the third argument is incorrect anyway.  */
+	  line_exchange(&line, &s_accum, true);
 	  if (line.length &&
 	      line.active[line.length - 1] == '\n')
 	    line.length--;
@@ -1210,12 +1315,12 @@ shrink_program(vec, cur_cmd)
   countT cmd_cnt;
 
   for (p=v; p < cur_cmd; ++p)
-    if (p->cmd != ':')
+    if (p->cmd != '#')
       MEMCPY(v++, p, sizeof *v);
   cmd_cnt = v - vec->v;
 
   for (; p < last_cmd; ++p)
-    if (p->cmd != ':')
+    if (p->cmd != '#')
       MEMCPY(v++, p, sizeof *v);
   vec->v_length = v - vec->v;
 
@@ -1238,7 +1343,7 @@ execute_program(vec, input)
   end_cmd = vec->v + vec->v_length;
   while (cur_cmd < end_cmd)
     {
-      if (match_address_p(cur_cmd, input))
+      if (match_address_p(cur_cmd, input) != cur_cmd->addr_bang)
 	{
 	  switch (cur_cmd->cmd)
 	    {
@@ -1256,12 +1361,13 @@ execute_program(vec, input)
 	      continue;
 
 	    case '}':
+	    case '#':
 	    case ':':
 	      /* Executing labels and block-ends are easy. */
 	      break;
 
 	    case 'c':
-	      if (!cur_cmd->a1_matched)
+	      if (cur_cmd->range_state != RANGE_ACTIVE)
 		output_line(cur_cmd->x.cmd_txt.text,
 			    cur_cmd->x.cmd_txt.text_length - 1, true,
 			    &output_file);
@@ -1291,9 +1397,7 @@ execute_program(vec, input)
 #ifdef HAVE_POPEN
 	      FILE *pipe;
 	      int cmd_length = cur_cmd->x.cmd_txt.text_length;
-	      if (s_accum.alloc == 0)
-		line_init(&s_accum, INITIAL_BUFFER_SIZE);
-	      s_accum.length = 0;
+	      line_reset(&s_accum, NULL);
 
 	      if (!cmd_length)
 		{
@@ -1309,16 +1413,16 @@ execute_program(vec, input)
 
 	      if (pipe != NULL) 
 		{
+		  char buf[4096];
+		  int n;
 		  while (!feof (pipe)) 
-		    {
-		      char buf[4096];
-		      int n = fread (buf, sizeof(char), 4096, pipe);
-		      if (n > 0)
+		    if ((n = fread (buf, sizeof(char), 4096, pipe)) > 0)
+		      {
 			if (!cmd_length)
 			  str_append(&s_accum, buf, n);
 			else
 			  ck_fwrite(buf, 1, n, output_file.fp);
-		    }
+		      }
 		  
 		  pclose (pipe);
 		  if (!cmd_length)
@@ -1330,8 +1434,9 @@ execute_program(vec, input)
 
 		      /* Exchange line and s_accum.  This can be much
 			 cheaper than copying s_accum.active into line.text
-			 (for huge lines). */
-		      line_exchange(&line, &s_accum);
+			 (for huge lines).  See comment above for 'g' as
+			 to while the third argument is incorrect anyway.  */
+		      line_exchange(&line, &s_accum, true);
 		    }
                   else
                     flush_output(output_file.fp);
@@ -1346,19 +1451,33 @@ execute_program(vec, input)
 	    }
 
 	    case 'g':
-	      line_copy(&hold, &line);
+	      /* We do not have a really good choice for the third parameter.
+		 The problem is that hold space and the input file might as
+		 well have different states; copying it from hold space means
+		 that subsequent input might be read incorrectly, while
+		 keeping it as in pattern space means that commands operating
+		 on the moved buffer might consider a wrong character set.
+		 We keep it true because it's what sed <= 4.1.5 did.  */
+	      line_copy(&hold, &line, true);
 	      break;
 
 	    case 'G':
-	      line_append(&hold, &line);
+	      /* We do not have a really good choice for the third parameter.
+		 The problem is that hold space and pattern space might as
+		 well have different states.  So, true is as wrong as false.
+		 We keep it true because it's what sed <= 4.1.5 did, but
+		 we could consider having line_ap.  */
+	      line_append(&hold, &line, true);
 	      break;
 
 	    case 'h':
-	      line_copy(&line, &hold);
+	      /* Here, it is ok to have true.  */
+	      line_copy(&line, &hold, true);
 	      break;
 
 	    case 'H':
-	      line_append(&line, &hold);
+	      /* See comment above for 'G' regarding the third parameter.  */
+	      line_append(&line, &hold, true);
 	      break;
 
 	    case 'i':
@@ -1418,6 +1537,7 @@ execute_program(vec, input)
             case 'q':
               if (!no_default_output)
                 output_line(line.active, line.length, line.chomped, &output_file);
+	      dump_append_queue();
 
 	    case 'Q':
 	      return cur_cmd->x.int_arg == -1 ? 0 : cur_cmd->x.int_arg;
@@ -1438,8 +1558,7 @@ execute_program(vec, input)
 		  char *text = NULL;
 		  int result;
 
-		  result = getline (&text, &buflen, cur_cmd->x.fp);
-
+		  result = ck_getline (&text, &buflen, cur_cmd->x.fp);
 		  if (result != EOF)
 		    {
 		      aq = next_append_slot();
@@ -1489,7 +1608,8 @@ execute_program(vec, input)
 	      break;
 
 	    case 'x':
-	      line_exchange(&line, &hold);
+	      /* See comment above for 'g' regarding the third parameter.  */
+	      line_exchange(&line, &hold, false);
 	      break;
 
 	    case 'y':
@@ -1580,44 +1700,47 @@ execute_program(vec, input)
 	      break;
 
 	    default:
-	      panic(_("INTERNAL ERROR: Bad cmd %c"), cur_cmd->cmd);
+	      panic("INTERNAL ERROR: Bad cmd %c", cur_cmd->cmd);
 	    }
 	}
 
 #ifdef EXPERIMENTAL_DASH_N_OPTIMIZATION
       /* If our top-level program consists solely of commands with
-       * ADDR_IS_NUM addresses then once we past the last mentioned
-       * line we should be able to quit if no_default_output is true,
-       * or otherwise quickly copy input to output.  Now whether this
-       * optimization is a win or not depends on how cheaply we can
-       * implement this for the cases where it doesn't help, as
-       * compared against how much time is saved.  One semantic
-       * difference (which I think is an improvement) is that *this*
-       * version will terminate after printing line two in the script
-       * "yes | sed -n 2p". 
-       *
-       * Don't use this when in-place editing is active, because line
-       * numbers restart each time then. */
+         ADDR_IS_NUM addresses then once we past the last mentioned
+         line we should be able to quit if no_default_output is true,
+         or otherwise quickly copy input to output.  Now whether this
+         optimization is a win or not depends on how cheaply we can
+         implement this for the cases where it doesn't help, as
+         compared against how much time is saved.  One semantic
+         difference (which I think is an improvement) is that *this*
+         version will terminate after printing line two in the script
+         "yes | sed -n 2p". 
+        
+         Don't use this when in-place editing is active, because line
+         numbers restart each time then. */
       else if (!separate_files)
 	{
-	  /* can we ever match again? */
-	  if (cur_cmd->a1->addr_type == ADDR_IS_NUM &&
-	      ((input->line_number < cur_cmd->a1->addr_number)
-	       != !cur_cmd->addr_bang))
+	  if (cur_cmd->a1->addr_type == ADDR_IS_NUM
+	      && (cur_cmd->a2
+		  ? cur_cmd->range_state == RANGE_CLOSED
+		  : cur_cmd->a1->addr_number < input->line_number))
 	    {
-	      /* skip all this next time */
+	      /* Skip this address next time */
+	      cur_cmd->addr_bang = !cur_cmd->addr_bang;
 	      cur_cmd->a1->addr_type = ADDR_IS_NULL;
-	      cur_cmd->addr_bang = true;
+	      if (cur_cmd->a2)
+		cur_cmd->a2->addr_type = ADDR_IS_NULL;
 
 	      /* can we make an optimization? */
-	      if (cur_cmd->cmd == 'b' || cur_cmd->cmd == 't'
-		  || cur_cmd->cmd == '{')
-		--branches;
-	      cur_cmd->cmd = ':';	/* replace with no-op */
-	      if (branches == 0)
+	      if (cur_cmd->addr_bang)
 		{
-		  /* whew!  all that just so that we can get to here! */
-		  cur_cmd = shrink_program(vec, cur_cmd);
+		  if (cur_cmd->cmd == 'b' || cur_cmd->cmd == 't'
+		      || cur_cmd->cmd == 'T' || cur_cmd->cmd == '}')
+		    branches--;
+
+		  cur_cmd->cmd = '#';	/* replace with no-op */
+	          if (branches == 0)
+		    cur_cmd = shrink_program(vec, cur_cmd);
 		  if (!cur_cmd && no_default_output)
 		    return 0;
 		  end_cmd = vec->v + vec->v_length;
@@ -1651,16 +1774,21 @@ process_files(the_program, argv)
   struct input input;
   int status;
 
-  line_init(&line, INITIAL_BUFFER_SIZE);
-  line_init(&hold, 0);
-  line_init(&buffer, 0);
+  line_init(&line, NULL, INITIAL_BUFFER_SIZE);
+  line_init(&hold, NULL, 0);
+  line_init(&buffer, NULL, 0);
 
 #ifdef EXPERIMENTAL_DASH_N_OPTIMIZATION
   branches = count_branches(the_program);
 #endif /*EXPERIMENTAL_DASH_N_OPTIMIZATION*/
-  input.file_list = stdin_argv;
+  input.reset_at_next_file = true;
   if (argv && *argv)
     input.file_list = argv;
+  else if (in_place_extension)
+    panic(_("no input files"));
+  else
+    input.file_list = stdin_argv;
+
   input.bad_count = 0;
   input.line_number = 0;
   input.read_fn = read_always_fail;
